﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using Harmony;
using RimWorld;
using UnityEngine;
using Verse;
using Verse.Sound;

namespace SlayerWeaponBase
{






    [StaticConstructorOnStartup]
    [HarmonyPatch(typeof(Pawn_MeleeVerbs))]
    [HarmonyPatch("ChooseMeleeVerb")]
    class Pawn_MeleeVerbs_Patch
    {
        /// <summary>
        /// 初回起動でハーモニの起動を行います。
        /// </summary>
        static Pawn_MeleeVerbs_Patch()
        {
            var harmony = HarmonyInstance.Create("com.github.harmony.rimworld.mod.SlayerWeaponBase");
            harmony.PatchAll(Assembly.GetExecutingAssembly());
        }

        static readonly Type[] m_SetCurMeleeVerbParamTypes  = new Type[] { typeof(Verb) };
        static readonly MethodInfo m_Method = AccessTools.Method(typeof(Pawn_MeleeVerbs), "SetCurMeleeVerb", m_SetCurMeleeVerbParamTypes);
        static readonly object[] m_Null = new object[] { null };
        static readonly object[] m_Verb = new object[] { null };

        static bool Prefix(Pawn_MeleeVerbs __instance)
        {
            bool slayer = false;
            Traverse traverseInstance = Traverse.Create(__instance);
            Pawn pawn = traverseInstance.Field("pawn").GetValue<Pawn>();

            if (pawn.equipment != null && pawn.equipment.Primary != null)
            {
                SlayerWeaponBase wb = pawn.equipment.Primary.def as SlayerWeaponBase;
                slayer = !(wb is null);
            }

            if (!slayer)
                return true;

            IEnumerable<VerbEntry> updatedAvailableVerbsList = (from x in __instance.GetUpdatedAvailableVerbsList() where x.verb is SWB_Verb_MeleeAttack select x);

            if (updatedAvailableVerbsList.IsEmpty())
            {
                Log.ErrorOnce(string.Format("{0} has no available melee attack", pawn), 1664289);
                m_Method.Invoke(__instance, m_Null);
            }
            else
            {
                m_Verb[0] = updatedAvailableVerbsList.RandomElementByWeight((VerbEntry ve) => ve.SelectionWeight).verb;
                m_Method.Invoke(__instance, m_Verb);
            }
            return false;
        }
    }

}
