﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;

using RimWorld;
using UnityEngine;
using Verse;
using Verse.Sound;

namespace SlayerWeaponBase
{
    public class SWB_Verb_MeleeAttack : SWB_Verb_MeleeAttackBase
    {

        protected override bool TryCastShot()
        {
            Pawn casterPawn = base.CasterPawn;
            if (casterPawn.stances.FullBodyBusy)
            {
                return false;
            }


            Thing thing = this.currentTarget.Thing;
            // modified-str
            if (casterPawn.equipment.Primary == null)
            {
                return false;
            }
            if (this.ownerEquipment == null)
            {
                return base.TryCastShot();
            }

            SlayerWeaponBase uniqueProperties = this.ownerEquipment.def as SlayerWeaponBase;
            if (uniqueProperties == null)
            {
                Log.Warning("SlayerMeleeWeaponBase property get method is missing... so call default method!");
                return base.TryCastShot();
            }
            List<SlayerSpecialMove> sspms = uniqueProperties.ml_specialMoves;
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationML.BEFORE_EVAL, sspms);

            bool isNormal = false;
            List<TraitSet> requireTraits = null;
            List<SecondarySkill> secondarySkills = null;
            if (m_InitConfigured)
            {
                requireTraits = m_Traits;
                secondarySkills = m_SecondarySkills;
            }
            else
            {
                List<TraitInfo> traitInfos = uniqueProperties.ml_traitInfos;
                if (traitInfos == null)
                {
                    traitInfos = new List<TraitInfo>();
                }
                if (uniqueProperties.ml_traitAliases != null)
                {
                    SlayerWeaponBase.AddTraitInfoByAlias(traitInfos, uniqueProperties.ml_traitAliases.ToArray());
                }
                requireTraits = new List<TraitSet>();
                foreach (TraitInfo info in traitInfos)
                {
                    TraitSet ts = new TraitSet { traitDef = TraitDef.Named(info.traitDefName), degree = info.degree };
                    if (ts.traitDef != null)
                    {
                        requireTraits.Add(ts);
                    }
                }
                m_Traits = requireTraits;

#if DEBUG
                Utility.AddDebug("ここまで大丈夫だよ！！＠１");
#endif
                secondarySkills = uniqueProperties.ml_secondarySkills;
                if (secondarySkills == null)
                {
                    secondarySkills = new List<SecondarySkill>();
                }
                m_SecondarySkills = secondarySkills;

                SetCooldown(uniqueProperties, 0.1f);
#if DEBUG
                Utility.AddDebug("ここまで大丈夫だよ！！＠２");
#endif

                m_Slayers = new Dictionary<string, float>();
                List<SlayerProperty> slayerProps = uniqueProperties.ml_slayerProps;
                if (slayerProps != null)
                {
                    foreach (SlayerProperty prop in slayerProps)
                    {
                        if (prop.targetDefName is null)
                            break;
                        m_Slayers[prop.targetDefName] = prop.slayerRatio;
                    }
                }
                m_InitConfigured = true;

            }


            m_skillBonusRatio = (int)uniqueProperties.ml_skillBonusRatio;
            m_ColonistOnly = uniqueProperties.ml_colonistOnly;

            foreach (TraitSet ts in requireTraits)
            {
                Trait trait = casterPawn.story.traits.GetTrait(ts.traitDef);
#if DEBUG
                Utility.AddDebug("trait={0}", trait);
                if (trait != null)
                    Utility.AddDebug("defName={0}, degree={1}", trait.def.defName, trait.Degree);
#endif
                if (!HasTrait(trait, ts.degree))
                {
                    isNormal = true;
                    break;
                }
            }

#if DEBUG
            Utility.AddDebug("isNormal={0}", isNormal);
#endif

            if (!isNormal)
            {
                if (m_ColonistOnly && !casterPawn.IsColonist)
                {
                    isNormal = true;
                }
            }

            m_AddAttackNum = (isNormal) ? 0 : (int)uniqueProperties.ml_addAttackNum;

#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠３");
#endif
            if (!isNormal)
            {
                if ((m_skillBonusRatio >= 1) && (m_skillBonusRatio <= 20))
                {
                    SkillRecord melee = casterPawn.skills.GetSkill(SkillDefOf.Melee);
                    if (!(melee is null))
                    {
                        int skillLevel = melee.Level;
                        foreach (SecondarySkill ss in secondarySkills)
                        {
                            SkillRecord rec = casterPawn.skills.GetSkill(ss.skill);
                            if (rec == null)
                                continue;
                            skillLevel += (int)(rec.Level * ss.ratio);
                        }
                        m_AddAttackNum += (int)(skillLevel / m_skillBonusRatio);
                    }
                }

            }
#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠４");
#endif

            Thing target = thing;

            float cooldownDefault = GetCooldownDefault();
            int maxCount = (int)(cooldownDefault / 0.1f);

            if (!isNormal)
            {
                maxCount += m_AddAttackNum;
            }

            m_AttackCountMax = maxCount;
            m_AddAttackNum++;//最低1回は攻撃

            m_AttackCount++;
            if (m_AttackCount > m_AddAttackNum)
            {
#if DEBUG
                Utility.AddDebug(String.Format("{0}、あなたは今待ち時間なので攻撃しちゃだめ。m_AttackCount={1}, m_AddAttackNum={2}, m_AttackCountMax={3}", casterPawn.Name, m_AttackCount, m_AddAttackNum, m_AttackCountMax));
#endif
                if (m_AttackCount > m_AttackCountMax)
                {
                    ResetCount();
                }

                return false;
            }

            bool slayerModified = !isNormal && m_Slayers.ContainsKey(target.def.defName);
            float damagePowerDefualt = this.tool.power;

            if (!isNormal && !(target.def.race is null))
            {
                if (slayerModified)
                {
                    this.tool.power *= m_Slayers[target.def.defName];
                }
                if (target.def.race != null)
                {
                    if (m_Slayers.ContainsKey(RACE_HUMANLIKE) && target.def.race.Humanlike)
                    {
                        this.tool.power *= m_Slayers[RACE_HUMANLIKE];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_ANIMAL) && target.def.race.Animal)
                    {
                        this.tool.power *= m_Slayers[RACE_ANIMAL];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_MECHANOID) && target.def.race.IsMechanoid)
                    {
                        this.tool.power *= m_Slayers[RACE_MECHANOID];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_INSECTOID) && target.def.race.FleshType == FleshTypeDefOf.Insectoid)
                    {
                        this.tool.power *= m_Slayers[RACE_INSECTOID];
                        slayerModified = true;
                    }
                }
            }
            // modified-end

            if (!base.CanHitTarget(thing))
            {
                Log.Warning(string.Concat(new object[]
                {
                    casterPawn,
                    " meleed ",
                    thing,
                    " from out of melee position."
                }));
            }
            casterPawn.rotationTracker.Face(thing.DrawPos);
#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠５");
#endif
            if (!this.IsTargetImmobile(this.currentTarget) && casterPawn.skills != null)
            {
                casterPawn.skills.Learn(SkillDefOf.Melee, 250f, false);
                foreach (SecondarySkill ss in secondarySkills)
                {
                    casterPawn.skills.Learn(ss.skill, ss.lernValue, false);
                }
            }
#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠６");
#endif

            // modified-str
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationML.BEFORE_DAM, sspms);
            // modified-end

            bool result;
            SoundDef soundDef;
            if (Rand.Value < this.GetNonMissChance(thing))
            {
                if (Rand.Value > this.GetDodgeChance(thing))
                {
                    BattleLogEntry_MeleeCombat log = this.CreateCombatLog(RulePackDefOf.Combat_Hit);
                    result = true;
                    this.ApplyMeleeDamageToTarget(this.currentTarget).InsertIntoLog(log);
                    if (thing.def.category == ThingCategory.Building)
                    {
                        soundDef = this.SoundHitBuilding();
                    }
                    else
                    {
                        soundDef = this.SoundHitPawn();
                    }
                }
                else
                {
                    result = false;
                    soundDef = this.SoundMiss();
                    MoteMaker.ThrowText(thing.DrawPos, thing.Map, "TextMote_Dodge".Translate(), 1.9f);
                    this.CreateCombatLog(RulePackDefOf.Combat_Dodge);
                }
            }
            else
            {
                result = false;
                soundDef = this.SoundMiss();
                this.CreateCombatLog(RulePackDefOf.Combat_Miss);
            }
            soundDef.PlayOneShot(new TargetInfo(thing.Position, casterPawn.Map, false));
            casterPawn.Drawer.Notify_MeleeAttackOn(thing);
            Pawn pawn = thing as Pawn;
            if (pawn != null && !pawn.Dead)
            {
                pawn.stances.StaggerFor(95);
                if (casterPawn.MentalStateDef != MentalStateDefOf.SocialFighting || pawn.MentalStateDef != MentalStateDefOf.SocialFighting)
                {
                    pawn.mindState.meleeThreat = casterPawn;
                    pawn.mindState.lastMeleeThreatHarmTick = Find.TickManager.TicksGame;
                }
            }
            casterPawn.rotationTracker.FaceCell(thing.Position);
            if (casterPawn.caller != null)
            {
                casterPawn.caller.Notify_DidMeleeAttack();
            }

            // modified-str
            //if (this.currentTarget.ThingDestroyed)
            //{
            //    ResetCount();
            //}
            //else
            //{
            //    Pawn tp = this.currentTarget.Thing as Pawn;
            //    if (tp != null && (tp.Dead || tp.Downed))
            //    {
            //        ResetCount();
            //    }
            //}
            if (slayerModified)
            {
                this.tool.power = damagePowerDefualt;
            }
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationML.BEFORE_END, sspms);
            // modified-end

            return result;
        }
    }





    public class SWB_Verb_RangedAttack : SWB_Verb_RangedAttackBase
    {

        protected override bool TryCastShotImpl()
        {
            // modified-str
            if (CasterPawn.equipment.Primary == null)
            {
                return false;
            }
            SlayerWeaponBase uniqueProperties = this.ownerEquipment.def as SlayerWeaponBase;
            if (uniqueProperties == null)
            {
                Log.Warning("SlayerRangedWeaponBase property get method is missing... so call default method!");
                return base.TryCastShotImpl();
            }
            List<SlayerSpecialMove> sspms = uniqueProperties.ml_specialMoves;
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_EVAL, sspms);
            // modified-end

            if (this.currentTarget.HasThing && this.currentTarget.Thing.Map != this.caster.Map)
            {
                return false;
            }
            ThingDef projectile = this.Projectile;
            if (projectile == null)
            {
                return false;
            }
            ShootLine shootLine;
            bool flag = base.TryFindShootLineFromTo(this.caster.Position, this.currentTarget, out shootLine);
            if (this.verbProps.stopBurstWithoutLos && !flag)
            {
                return false;
            }

            // modified-str
            Pawn casterPawn = this.CasterPawn;
            if (casterPawn.equipment.Primary == null)
            {
                return false;
            }
            m_skillBonusRatio = (int)uniqueProperties.rg_skillBonusRatio;
            m_ColonistOnly = uniqueProperties.rg_colonistOnly;

            bool isNormal = false;
            List<TraitSet> requireTraits = null;
            List<SecondarySkill> secondarySkills = null;
            if (m_InitConfigured)
            {
                requireTraits = m_Traits;
                secondarySkills = m_SecondarySkills;
            }
            else
            {
                List<TraitInfo> traitInfos = uniqueProperties.rg_traitInfos;
                if (traitInfos == null)
                {
                    traitInfos = new List<TraitInfo>();
                }
                if (uniqueProperties.rg_traitAliases != null)
                {
                    SlayerWeaponBase.AddTraitInfoByAlias(traitInfos, uniqueProperties.rg_traitAliases.ToArray());
                }
                requireTraits = new List<TraitSet>();
                foreach (TraitInfo info in traitInfos)
                {
                    TraitSet ts = new TraitSet { traitDef = TraitDef.Named(info.traitDefName), degree = info.degree };
                    if (ts.traitDef != null)
                    {
                        requireTraits.Add(ts);
                    }
                }
                m_Traits = requireTraits;

                secondarySkills = uniqueProperties.rg_secondarySkills;
                if (secondarySkills == null)
                {
                    secondarySkills = new List<SecondarySkill>();
                }
                m_SecondarySkills = secondarySkills;

                SetCooldown(uniqueProperties, 0.1f);

                m_Slayers = new Dictionary<string, float>();
                List<SlayerProperty> slayerProps = uniqueProperties.rg_slayerProps;
                if (slayerProps != null)
                {
                    foreach (SlayerProperty prop in slayerProps)
                    {
                        if (prop.targetDefName is null)
                            break;
                        m_Slayers[prop.targetDefName] = prop.slayerRatio;
                    }
                }

                ThingDef wpDef = casterPawn.equipment.Primary.def;
                for (int i = 0; i < wpDef.statBases.Count; i++)
                {
                    if (wpDef.statBases[i].stat.defName == PROP_BASE_RGW_COOLDOWN)
                    {
                        float coolDown = 0.1f;
                        SetCooldownDefault(wpDef.statBases[i].value);
                        wpDef.statBases[i].value = coolDown;
                    }
                }
                m_BurstShotCountDefault = Math.Max(0, this.verbProps.burstShotCount - 1);
                this.verbProps.burstShotCount = 0;
                this.verbProps.warmupTime = 0;

                m_InitConfigured = true;

            }

            m_skillBonusRatio = (int)uniqueProperties.rg_skillBonusRatio;
            m_ColonistOnly = uniqueProperties.rg_colonistOnly;

            foreach (TraitSet ts in requireTraits)
            {
                Trait trait = casterPawn.story.traits.GetTrait(ts.traitDef);
#if DEBUG
                Utility.AddDebug("trait={0}", trait);
                if (trait != null)
                    Utility.AddDebug("defName={0}, degree={1}", trait.def.defName, trait.Degree);
#endif
                if (!HasTrait(trait, ts.degree))
                {
                    isNormal = true;
                    break;
                }
            }

#if DEBUG
            Utility.AddDebug("isNormal={0}", isNormal);
#endif

            if (!isNormal)
            {
                if (m_ColonistOnly && !casterPawn.IsColonist)
                {
                    isNormal = true;
                }
            }

            Thing target = this.currentTarget.Thing;

#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠１");
#endif

            int addAttackSkillNum = 0;
            if ((m_skillBonusRatio >= 1) && (m_skillBonusRatio <= 20))
            {
                SkillRecord shoothing = casterPawn.skills.GetSkill(SkillDefOf.Shooting);
                if (shoothing != null)
                {
                    int skillLvl = shoothing.Level;
                    foreach (SecondarySkill ss in secondarySkills)
                    {
                        SkillRecord rec = casterPawn.skills.GetSkill(ss.skill);
                        if (rec == null)
                            continue;
                        skillLvl += (int)(rec.Level * ss.ratio);

                    }
                    addAttackSkillNum += (int)(skillLvl / m_skillBonusRatio);
                }
            }
            m_AddAttackNum = (isNormal) ? 0 : (int)uniqueProperties.rg_addAttackNum + addAttackSkillNum + m_BurstShotCountDefault;

#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠２");
#endif
            float cooldownDefault = GetCooldownDefault();
            int countByCooldownRatio = (int)(cooldownDefault / 0.1f);
            int maxCount = countByCooldownRatio + m_BurstShotCountDefault;
            if (!isNormal)
            {
                maxCount = countByCooldownRatio + m_AddAttackNum;
            }
            m_AttackCountMax = maxCount;
            m_AddAttackNum++;//最低1回は攻撃

            CasterPos pos = m_PrevPos;
            int posX = this.caster.Position.x;
            int posY = this.caster.Position.y;
            int posZ = this.caster.Position.z;
            if (pos.x != posX || pos.y != posY || pos.z != posZ)
            {
                m_AttackCount = m_AttackCountMax;
                m_PrevPos.x = posX;
                m_PrevPos.y = posY;
                m_PrevPos.z = posZ;
            }

            m_AttackCount--;
            if (m_AttackCount >= m_AddAttackNum)
            {
                //AddDebug(String.Format("{0}、あなたは今待ち時間なので攻撃しちゃだめ。m_AttackCount={1}, m_AddAttackNum={2}, m_AttackCountMax={3}", casterPawn.Name, m_AttackCount, m_AddAttackNum, m_AttackCountMax));
                return false;
            }
            if (m_AttackCount < 0)
            {
                m_AttackCount = m_AttackCountMax;
                return false;
            }
            //AddDebug(String.Format("{0}、攻撃していいよ！。m_AttackCount={1}, m_AddAttackNum={2}, m_AttackCountMax={3}", casterPawn.Name, m_AttackCount, m_AddAttackNum, m_AttackCountMax));

            bool slayerModified = !isNormal && m_Slayers.ContainsKey(target.def.defName);
            float hitDamageGain = 1f;

            if (!isNormal && !(target.def.race is null))
            {
                if (slayerModified)
                {
                    hitDamageGain *= m_Slayers[target.def.defName];
                }
                if (target.def.race != null)
                {
                    if (m_Slayers.ContainsKey(RACE_HUMANLIKE) && target.def.race.Humanlike)
                    {
                        hitDamageGain *= m_Slayers[RACE_HUMANLIKE];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_ANIMAL) && target.def.race.Animal)
                    {
                        hitDamageGain *= m_Slayers[RACE_ANIMAL];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_MECHANOID) && target.def.race.IsMechanoid)
                    {
                        hitDamageGain *= m_Slayers[RACE_MECHANOID];
                        slayerModified = true;
                    }
                    if (m_Slayers.ContainsKey(RACE_INSECTOID) && target.def.race.FleshType == FleshTypeDefOf.Insectoid)
                    {
                        hitDamageGain *= m_Slayers[RACE_INSECTOID];
                        slayerModified = true;
                    }
                }
            }

#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠３");
#endif
            if (!this.IsTargetImmobile(this.currentTarget) && casterPawn.skills != null)
            {
                foreach (SecondarySkill ss in secondarySkills)
                {
                    casterPawn.skills.Learn(ss.skill, ss.lernValue, false);
                }
            }
#if DEBUG
            Utility.AddDebug("ここまで大丈夫だよ！！＠４");
#endif

            // modified-end

            if (this.ownerEquipment != null)
            {
                CompChangeableProjectile comp = this.ownerEquipment.GetComp<CompChangeableProjectile>();
                if (comp != null)
                {
                    comp.Notify_ProjectileLaunched();
                }
            } 
            Thing launcher = this.caster;
            Thing equipment = this.ownerEquipment;
            CompMannable compMannable = this.caster.TryGetComp<CompMannable>();
            if (compMannable != null && compMannable.ManningPawn != null)
            {
                launcher = compMannable.ManningPawn;
                equipment = this.caster;
            }
            Vector3 drawPos = this.caster.DrawPos;

            // modified - str
            ThingDef newProjectileDef = GetClonedProjectileDef(this.verbProps.defaultProjectile);
            newProjectileDef.projectile.damageAmountBase = (int)(newProjectileDef.projectile.damageAmountBase * hitDamageGain);
            Projectile projectile2 = (Projectile)GenSpawn.Spawn(newProjectileDef, shootLine.Source, this.caster.Map);
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_DAM, sspms);
            // modified - end


            projectile2.FreeIntercept = (this.canFreeInterceptNow && !projectile2.def.projectile.flyOverhead);
            if (this.verbProps.forcedMissRadius > 0.5f)
            {
                float num = (float)(this.currentTarget.Cell - this.caster.Position).LengthHorizontalSquared;
                float num2;
                if (num < 9f)
                {
                    num2 = 0f;
                }
                else if (num < 25f)
                {
                    num2 = this.verbProps.forcedMissRadius * 0.5f;
                }
                else if (num < 49f)
                {
                    num2 = this.verbProps.forcedMissRadius * 0.8f;
                }
                else
                {
                    num2 = this.verbProps.forcedMissRadius * 1f;
                }
                if (num2 > 0.5f)
                {
                    int max = GenRadial.NumCellsInRadius(this.verbProps.forcedMissRadius);
                    int num3 = Rand.Range(0, max);
                    if (num3 > 0)
                    {
                        if (DebugViewSettings.drawShooting)
                        {
                            MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToForRad", -1f);
                        }
                        IntVec3 c = this.currentTarget.Cell + GenRadial.RadialPattern[num3];
                        if (this.currentTarget.HasThing)
                        {
                            projectile2.ThingToNeverIntercept = this.currentTarget.Thing;
                        }
                        if (!projectile2.def.projectile.flyOverhead)
                        {
                            projectile2.InterceptWalls = true;
                        }
                        projectile2.Launch(launcher, drawPos, c, equipment, this.currentTarget.Thing);
                        // modified-str
                        SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_END, sspms);
                        // modified-end
                        return true;
                    }
                }
            }
            ShotReport shotReport = ShotReport.HitReportFor(this.caster, this, this.currentTarget);
            if (Rand.Value > shotReport.ChanceToNotGoWild_IgnoringPosture)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToWild", -1f);
                }
                shootLine.ChangeDestToMissWild();
                if (this.currentTarget.HasThing)
                {
                    projectile2.ThingToNeverIntercept = this.currentTarget.Thing;
                }
                if (!projectile2.def.projectile.flyOverhead)
                {
                    projectile2.InterceptWalls = true;
                }
                projectile2.Launch(launcher, drawPos, shootLine.Dest, equipment, this.currentTarget.Thing);
                // modified-str
                SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_END, sspms);
                // modified-end
                return true;
            }
            if (Rand.Value > shotReport.ChanceToNotHitCover)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToCover", -1f);
                }
                if (this.currentTarget.Thing != null && this.currentTarget.Thing.def.category == ThingCategory.Pawn)
                {
                    Thing randomCoverToMissInto = shotReport.GetRandomCoverToMissInto();
                    if (!projectile2.def.projectile.flyOverhead)
                    {
                        projectile2.InterceptWalls = true;
                    }
                    projectile2.Launch(launcher, drawPos, randomCoverToMissInto, equipment, this.currentTarget.Thing);
                    // modified-str
                    SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_END, sspms);
                    // modified-end
                    return true;
                }
            }
            if (DebugViewSettings.drawShooting)
            {
                MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToHit", -1f);
            }
            if (!projectile2.def.projectile.flyOverhead)
            {
                projectile2.InterceptWalls = (!this.currentTarget.HasThing || this.currentTarget.Thing.def.Fillage == FillCategory.Full);
            }
            if (this.currentTarget.Thing != null)
            {
                projectile2.Launch(launcher, drawPos, this.currentTarget, equipment, this.currentTarget.Thing);
            }
            else
            {
                projectile2.Launch(launcher, drawPos, shootLine.Dest, equipment, this.currentTarget.Thing);
            }
            // modified-str
            SpecialMoveController.DoSpecialMove(this, SpecialMoveLocationRG.BEFORE_END, sspms);
            // modified-end
            return true;
        }

    }

}
