﻿using RimWorld;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using System.Linq;
using UnityEngine;
using Verse;

namespace Verb_Shoot_MLS
{
    public class Verb_LaunchProjectile_MLS : Verb
    {
        public Verb_LaunchProjectile_MLS()
        {
        }

        public virtual ThingDef Projectile
        {
            get
            {
                if (this.ownerEquipment != null)
                {
                    CompChangeableProjectile comp = this.ownerEquipment.GetComp<CompChangeableProjectile>();
                    if (comp != null && comp.Loaded)
                    {
                        return comp.Projectile;
                    }
                }
                return this.verbProps.defaultProjectile;
            }
        }

        public override float HighlightFieldRadiusAroundTarget(out bool needLOSToCenter)
        {
            needLOSToCenter = true;
            ThingDef projectile = this.Projectile;
            if (projectile == null)
            {
                return 0f;
            }
            return projectile.projectile.explosionRadius;
        }

        ///// マルチロック用複射
        private bool TryCastShotSubTerget(LocalTargetInfo subterget,ThingDef subprojectileDef)
        {
            ShootLine shootLine;
            float single;
            if (subterget.HasThing && subterget.Thing.Map != this.caster.Map)
            {
                return false;
            }
            ThingDef projectile = subprojectileDef;
            if (projectile == null)
            {
                return false;
            }
            bool flag = base.TryFindShootLineFromTo(this.caster.Position, subterget, out shootLine);
            if (this.verbProps.stopBurstWithoutLos && !flag)
            {
                return false;
            }
            if (this.ownerEquipment != null)
            {
                CompChangeableProjectile comp = this.ownerEquipment.GetComp<CompChangeableProjectile>();
                if (comp != null)
                {
                    comp.Notify_ProjectileLaunched();
                }
            }
            Thing manningPawn = this.caster;
            Thing thing = this.ownerEquipment;
            CompMannable compMannable = this.caster.TryGetComp<CompMannable>();
            if (compMannable != null && compMannable.ManningPawn != null)
            {
                manningPawn = compMannable.ManningPawn;
                thing = this.caster;
            }
            Vector3 drawPos = this.caster.DrawPos;
            Projectile thing1 = (Projectile)GenSpawn.Spawn(projectile, shootLine.Source, this.caster.Map);
            thing1.FreeIntercept = (!this.canFreeInterceptNow ? false : !thing1.def.projectile.flyOverhead);
            if (this.verbProps.forcedMissRadius > 0.5f)
            {
                IntVec3 cell = subterget.Cell - this.caster.Position;
                float lengthHorizontalSquared = (float)cell.LengthHorizontalSquared;
                if (lengthHorizontalSquared < 9f)
                {
                    single = 0f;
                }
                else if (lengthHorizontalSquared >= 25f)
                {
                    single = (lengthHorizontalSquared >= 49f ? this.verbProps.forcedMissRadius * 1f : this.verbProps.forcedMissRadius * 0.8f);
                }
                else
                {
                    single = this.verbProps.forcedMissRadius * 0.5f;
                }
                if (single > 0.5f)
                {
                    int num = GenRadial.NumCellsInRadius(this.verbProps.forcedMissRadius);
                    int num1 = Rand.Range(0, num);
                    if (num1 > 0)
                    {
                        if (DebugViewSettings.drawShooting)
                        {
                            MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToForRad", -1f);
                        }
                        IntVec3 intVec3 = subterget.Cell + GenRadial.RadialPattern[num1];
                        if (subterget.HasThing)
                        {
                            thing1.ThingToNeverIntercept = subterget.Thing;
                        }
                        if (!thing1.def.projectile.flyOverhead)
                        {
                            thing1.InterceptWalls = true;
                        }
                        thing1.Launch(manningPawn, drawPos, intVec3, thing, subterget.Thing);
                        return true;
                    }
                }
            }
            ShotReport shotReport = ShotReport.HitReportFor(this.caster, this, subterget);
            if (Rand.Value > shotReport.ChanceToNotGoWild_IgnoringPosture)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToWild", -1f);
                }
                shootLine.ChangeDestToMissWild();
                if (subterget.HasThing)
                {
                    thing1.ThingToNeverIntercept = subterget.Thing;
                }
                if (!thing1.def.projectile.flyOverhead)
                {
                    thing1.InterceptWalls = true;
                }
                thing1.Launch(manningPawn, drawPos, shootLine.Dest, thing, subterget.Thing);
                return true;
            }
            if (Rand.Value > shotReport.ChanceToNotHitCover)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToCover", -1f);
                }
                if (subterget.Thing != null && subterget.Thing.def.category == ThingCategory.Pawn)
                {
                    Thing randomCoverToMissInto = shotReport.GetRandomCoverToMissInto();
                    if (!thing1.def.projectile.flyOverhead)
                    {
                        thing1.InterceptWalls = true;
                    }
                    thing1.Launch(manningPawn, drawPos, randomCoverToMissInto, thing, subterget.Thing);
                    return true;
                }
            }
            if (DebugViewSettings.drawShooting)
            {
                MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToHit", -1f);
            }
            if (!thing1.def.projectile.flyOverhead)
            {
                thing1.InterceptWalls = (!subterget.HasThing ? true : subterget.Thing.def.Fillage == FillCategory.Full);
            }
            if (subterget.Thing == null)
            {
                thing1.Launch(manningPawn, drawPos, shootLine.Dest, thing, subterget.Thing);
            }
            else
            {
                thing1.Launch(manningPawn, drawPos, subterget, thing, subterget.Thing);
            }
            return true;
        }

        protected override bool TryCastShot()
        {
            ShootLine shootLine;
            float single;
            if (this.currentTarget.HasThing && this.currentTarget.Thing.Map != this.caster.Map)
            {
                return false;
            }
            ThingDef projectile = this.Projectile;
            if (projectile == null)
            {
                return false;
            }
            bool flag = base.TryFindShootLineFromTo(this.caster.Position, this.currentTarget, out shootLine);
            if (this.verbProps.stopBurstWithoutLos && !flag)
            {
                return false;
            }
            if (this.ownerEquipment != null)
            {
                CompChangeableProjectile comp = this.ownerEquipment.GetComp<CompChangeableProjectile>();
                if (comp != null)
                {
                    comp.Notify_ProjectileLaunched();
                }
            }
            Thing manningPawn = this.caster;
            Thing thing = this.ownerEquipment;
            CompMannable compMannable = this.caster.TryGetComp<CompMannable>();
            if (compMannable != null && compMannable.ManningPawn != null)
            {
                manningPawn = compMannable.ManningPawn;
                thing = this.caster;
            }
            /////////////////マルチロックSTART////////////////////////////////
            List<LocalTargetInfo> targetlist = new List<LocalTargetInfo>();
            Pawn targetpawn;
            bool flgtmp = true;
            //ターゲットがPawnの場合に限定
            if (this.currentTarget.Thing != null && this.currentTarget.Thing.def.category == ThingCategory.Pawn)
            {
                targetpawn = this.currentTarget.Thing as Pawn;
                //ターゲットが入植者以外に限定
                if (targetpawn != null && !targetpawn.IsColonist)
                {
                    IEnumerable<Pawn> subtargets =
                        from st in targetpawn.Map.mapPawns.AllPawns
                        where st.Spawned && !st.Dead && !st.Downed && st != targetpawn
                        select st;
                    if (subtargets.Any<Pawn>())
                    {
                        foreach(Pawn st in subtargets)
                        {
                            //メインターゲットが勢力あり
                            if(targetpawn.Faction != null)
                            {
                                //サブターゲットと同一勢力、囚人の場合は脱走中以外が対象
                                if (st.Faction == targetpawn.Faction && (!st.IsPrisoner || PrisonBreakUtility.IsPrisonBreaking(st)))
                                {
                                    if (st.Position.InHorDistOf(targetpawn.Position, 15f))
                                    {
                                        flgtmp = TryCastShotSubTerget(new LocalTargetInfo((Thing)st), projectile);
                                    }
                                }
                            }
                            //メインターゲットが勢力なし
                            else
                            {
                                //サブターゲットが人狩り
                                if (st.mindState != null && st.mindState.mentalStateHandler.CurState != null
                                    && (st.mindState.mentalStateHandler.CurStateDef == MentalStateDefOf.Manhunter
                                        || st.mindState.mentalStateHandler.CurStateDef == MentalStateDefOf.ManhunterPermanent)
                                    )
                                {
                                    if (st.Position.InHorDistOf(targetpawn.Position, 15f))
                                    {
                                        flgtmp = TryCastShotSubTerget(new LocalTargetInfo((Thing)st), projectile);
                                    }
                                }
                            }
                        }
                    }
                }
            }
            /////////////////マルチロックEND////////////////////////////////


            Vector3 drawPos = this.caster.DrawPos;
            Projectile thing1 = (Projectile)GenSpawn.Spawn(projectile, shootLine.Source, this.caster.Map);
            thing1.FreeIntercept = (!this.canFreeInterceptNow ? false : !thing1.def.projectile.flyOverhead);
            if (this.verbProps.forcedMissRadius > 0.5f)
            {
                IntVec3 cell = this.currentTarget.Cell - this.caster.Position;
                float lengthHorizontalSquared = (float)cell.LengthHorizontalSquared;
                if (lengthHorizontalSquared < 9f)
                {
                    single = 0f;
                }
                else if (lengthHorizontalSquared >= 25f)
                {
                    single = (lengthHorizontalSquared >= 49f ? this.verbProps.forcedMissRadius * 1f : this.verbProps.forcedMissRadius * 0.8f);
                }
                else
                {
                    single = this.verbProps.forcedMissRadius * 0.5f;
                }
                if (single > 0.5f)
                {
                    int num = GenRadial.NumCellsInRadius(this.verbProps.forcedMissRadius);
                    int num1 = Rand.Range(0, num);
                    if (num1 > 0)
                    {
                        if (DebugViewSettings.drawShooting)
                        {
                            MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToForRad", -1f);
                        }
                        IntVec3 intVec3 = this.currentTarget.Cell + GenRadial.RadialPattern[num1];
                        if (this.currentTarget.HasThing)
                        {
                            thing1.ThingToNeverIntercept = this.currentTarget.Thing;
                        }
                        if (!thing1.def.projectile.flyOverhead)
                        {
                            thing1.InterceptWalls = true;
                        }
                        thing1.Launch(manningPawn, drawPos, intVec3, thing, this.currentTarget.Thing);
                        return true;
                    }
                }
            }
            ShotReport shotReport = ShotReport.HitReportFor(this.caster, this, this.currentTarget);
            if (Rand.Value > shotReport.ChanceToNotGoWild_IgnoringPosture)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToWild", -1f);
                }
                shootLine.ChangeDestToMissWild();
                if (this.currentTarget.HasThing)
                {
                    thing1.ThingToNeverIntercept = this.currentTarget.Thing;
                }
                if (!thing1.def.projectile.flyOverhead)
                {
                    thing1.InterceptWalls = true;
                }
                thing1.Launch(manningPawn, drawPos, shootLine.Dest, thing, this.currentTarget.Thing);
                return true;
            }
            if (Rand.Value > shotReport.ChanceToNotHitCover)
            {
                if (DebugViewSettings.drawShooting)
                {
                    MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToCover", -1f);
                }
                if (this.currentTarget.Thing != null && this.currentTarget.Thing.def.category == ThingCategory.Pawn)
                {
                    Thing randomCoverToMissInto = shotReport.GetRandomCoverToMissInto();
                    if (!thing1.def.projectile.flyOverhead)
                    {
                        thing1.InterceptWalls = true;
                    }
                    thing1.Launch(manningPawn, drawPos, randomCoverToMissInto, thing, this.currentTarget.Thing);
                    return true;
                }
            }
            if (DebugViewSettings.drawShooting)
            {
                MoteMaker.ThrowText(this.caster.DrawPos, this.caster.Map, "ToHit", -1f);
            }
            if (!thing1.def.projectile.flyOverhead)
            {
                thing1.InterceptWalls = (!this.currentTarget.HasThing ? true : this.currentTarget.Thing.def.Fillage == FillCategory.Full);
            }
            if (this.currentTarget.Thing == null)
            {
                thing1.Launch(manningPawn, drawPos, shootLine.Dest, thing, this.currentTarget.Thing);
            }
            else
            {
                thing1.Launch(manningPawn, drawPos, this.currentTarget, thing, this.currentTarget.Thing);
            }
            return true;
        }

        public override void WarmupComplete()
        {
            Thing thing;
            ThingDef thingDef;
            base.WarmupComplete();
            BattleLog battleLog = Find.BattleLog;
            Thing thing1 = this.caster;
            if (!this.currentTarget.HasThing)
            {
                thing = null;
            }
            else
            {
                thing = this.currentTarget.Thing;
            }
            if (this.ownerEquipment == null)
            {
                thingDef = null;
            }
            else
            {
                thingDef = this.ownerEquipment.def;
            }
            battleLog.Add(new BattleLogEntry_RangedFire(thing1, thing, thingDef, this.Projectile, this.ShotsPerBurst > 1));
        }

    }


    /* こっから射撃 **********************************************************************/
    public class Verb_Shoot_MLS : Verb_LaunchProjectile_MLS
    {
        protected override int ShotsPerBurst
        {
            get
            {
                return this.verbProps.burstShotCount;
            }
        }

        public Verb_Shoot_MLS()
        {
        }

        protected override bool TryCastShot()
        {
            bool flag = base.TryCastShot();
            if (flag && base.CasterIsPawn)
            {
                base.CasterPawn.records.Increment(RecordDefOf.ShotsFired);
            }
            return flag;
        }

        public override void WarmupComplete()
        {
            base.WarmupComplete();
            if (base.CasterIsPawn && base.CasterPawn.skills != null)
            {
                float single = 6f;
                Pawn thing = this.currentTarget.Thing as Pawn;
                if (thing != null && thing.HostileTo(this.caster) && !thing.Downed)
                {
                    single = 240f;
                }
                base.CasterPawn.skills.Learn(SkillDefOf.Shooting, single, false);
            }
        }
    }
}
